#ifndef NUMPY_CORE_SRC_MULTIARRAY_TEXTREADING_READTEXT_H_
#define NUMPY_CORE_SRC_MULTIARRAY_TEXTREADING_READTEXT_H_

NPY_NO_EXPORT PyObject *
_load_from_filelike(PyObject *mod,
        PyObject *const *args, Py_ssize_t len_args, PyObject *kwnames);

#endif  /* NUMPY_CORE_SRC_MULTIARRAY_TEXTREADING_READTEXT_H_ */
