#ifndef NUMPY_CORE_SRC_MULTIARRAY_SEQUENCE_H_
#define NUMPY_CORE_SRC_MULTIARRAY_SEQUENCE_H_

extern NPY_NO_EXPORT PySequenceMethods array_as_sequence;

#endif  /* NUMPY_CORE_SRC_MULTIARRAY_SEQUENCE_H_ */
